 /*
  * client.c
  * 
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
  * Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  * 
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  * See the GNU General Public License for more details.
  * 
  * You should have received a copy of the GNU General Public License along
  * with this program.  If not, see <http://www.gnu.org/licenses/>.
  * 
  * See the COPYING file.
  */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libubus.h>
#include <libubox/blobmsg_json.h>

/*
==================================================================================
referred to sysexits.h as suggested by marc<marcxdv@welz.org.za>

* https://www.freebsd.org/cgi/man.cgi?query=sysexits&apropos=0&sektion=0&manpath=FreeBSD+4.3-RELEASE&format=html

Since there are error return values that have a defined meaning, it
makes sense to use them as they are meant to.

 Error values not defined here, have the meanings of the external
 program's return value.

 return value of 101 means operation not defined
 return value of 64 means incorrect number of parameter for operation selected
 return value of 65 means incorrect input data in some way
==================================================================================
*/

#define ERR_OP_NOT_DEFINED 101
#define ERR_INVALID_NO_OF_PARAMS 64
#define ERR_INVALID_DATA 65


typedef enum _op_t {
	opNone = -1, // must be first/lowest!
	opMount = 0,
	opUmount = 1,
	opRemove = 2,
	opUndefined = 3 // must be last/highest!
} op_t;

 
static struct ubus_context *ctx;
static struct blob_buf bb_t;
 
static void result_handler(struct ubus_request *req, int type, struct blob_attr *msg);

static int util_split(const char *str, char c, char ***array)
{
    int count = 1;
    int token_len = 1;
    int i = 0;
    char *t;

    const char *p = str;
    while (*p != '\0') {
        if (*p == c)
            count++;
        p++;
    }

    *array = (char**) malloc(sizeof(char*) * count);
    if (*array == NULL) {
		return -ENOMEM;		
	}

    p = str;
    while (*p != '\0')
    {
        if (*p == c) {
            (*array)[i] = (char*) malloc( sizeof(char) * token_len );
            if ((*array)[i] == NULL) {
				return -ENOMEM;
			}

            token_len = 0;
            i++;
        }
        p++;
        token_len++;
    }
    
    (*array)[i] = (char*) malloc( sizeof(char) * token_len );
    if ((*array)[i] == NULL) {
		return -ENOMEM;
	}

    i = 0;
    p = str;
    t = ((*array)[i]);
    while (*p != '\0') {
        if (*p != c && *p != '\0') {
            *t = *p;
            t++;
        }
        else {
            *t = '\0';
            i++;
            t = ((*array)[i]);
        }
        p++;
    }

    return count;
}
 
/* ubus object assignment */
static struct ubus_object cli_obj=
{
	.name = "dmount-client",
	#if 0
	.type = &cli_obj_type,
	.methods = cli_methods,
	.n_methods = ARRAY_SIZE(cli_methods),
	.path= /* useless */
	#endif
};
 
 
int main(int argc, char **argv)
{
	op_t switch_item = opUndefined;
	char *endptr;
	long int res;
	int num_params_valid;
    int c2 = 0, c3 = 0;
    char **arr2 = NULL;
    char **arr3 = NULL;
	
	int ret;
	uint32_t host_id; /* ering host id */
	const char *ubus_socket=NULL; /* use default UNIX sock path: /var/run/ubus.sock */

	if(argc <= 1) {
		fprintf(
			stderr, "ERROR: " 
			"main(): No parameters supplied (try --help).\n"
		);

		return ERR_OP_NOT_DEFINED;
	}

	if(
		0==strcmp(argv[1], "-h") ||
		0==strcmp(argv[1], "--help")
	) {
		printf(
			"Usage:\n"
			"   dmount -h|--help        # Print this message.\n"
			"   dmount 0 DEVICE TARGET  # Mount DEVICE in the specified TARGET.\n"
			"   dmount 1 DEVICE         # Unmount DEVICE.\n"
			"   dmount 2 DEVICE         # Remove DEVICE\n"
		);

		exit(0);
	}

	// check conversion status (endptr) and converted value (res)

	res = strtol(argv[1], &endptr, 3);

	if(
		*endptr!='\0' ||
		res <= opNone ||
		res >= opUndefined
	) {
		fprintf(stderr, "Operation not defined\n");

		return ERR_OP_NOT_DEFINED;
	}

	switch_item = (int) res;

	// check for valid arguments

	num_params_valid = 0;

    if (argc == 3) {
		
		if(
			switch_item == opUmount ||
			switch_item == opRemove
		)
			num_params_valid = 1;

		if (!num_params_valid) {
			fprintf(stderr, "Incorrect number of parameters for operation selected.\n");

			return ERR_INVALID_NO_OF_PARAMS;
		}

        c2 = util_split(argv[2], '/', &arr2);
        if(c2 != 3) {			
			fprintf(stderr, "Wrong device name. You should specify the full path, e.g. /dev/sdb1");

			return ERR_INVALID_DATA;
		}
		
		if(strcmp(arr2[1], "dev") && switch_item == opUmount) {			
			fprintf(stderr, "Wrong device name. You should specify the full path, e.g. /dev/sdb1");

			return ERR_INVALID_DATA;
		}
		
		if(strcmp(arr2[1], "dev") && switch_item == opRemove) {			
			fprintf(stderr, "Wrong device name. You should specify the full path, e.g. /dev/sdb");

			return ERR_INVALID_DATA;
		}
	}
	else if (argc == 4) {
		if( switch_item == opMount )
			num_params_valid = 1;

		if (!num_params_valid) {
			fprintf(stderr, "Incorrect number of parameters for operation selected.\n");

			return ERR_INVALID_NO_OF_PARAMS;
		}

        c2 = util_split(argv[2], '/', &arr2);
        if(c2 != 3) {			
			fprintf(stderr, "Wrong device name. You should specify the full path, e.g. /dev/sdb1");

			return ERR_INVALID_DATA;
		}
		
		if(strcmp(arr2[1], "dev")) {			
			fprintf(stderr, "Wrong device name. You should specify the full path, e.g. /dev/sdb1");

			return ERR_INVALID_DATA;
		}

        c3 = util_split(argv[3], '/', &arr3);
        if(c3 > 3) {			
			fprintf(stderr, "Wrong mountpoint.");

			return ERR_INVALID_DATA;
		}
		
		if(strcmp(arr3[1], "media")) {			
			fprintf(stderr, "Wrong mountpoint. You should specify some directory in /media");

			return ERR_INVALID_DATA;
		}
	}
	else {
		fprintf(stderr, "Incorrect number of parameters for operation selected.\n");

		return ERR_INVALID_NO_OF_PARAMS;
	}
 
	/* 1. create an epoll instance descriptor poll_fd */
	uloop_init();
 
	/* 2. connect to ubusd and get the context ctx */
	ctx=ubus_connect(ubus_socket);
	if(ctx==NULL) {
		printf("Fail to connect to ubusd!\n");
		return 1;
	}
 
	/* 3. registger epoll events to uloop, start sock listening */
	ubus_add_uloop(ctx);
 
	/* 4. register a usb_object to ubusd */
	ret=ubus_add_object(ctx, &cli_obj);
	if(ret!=0) {
		printf("Fail to register an object to ubus.\n");
		goto UBUS_FAIL;
 
	} else {
		printf("Add '%s' to ubus @%u successfully.\n",cli_obj.name, cli_obj.id);
	}
 
	/* 5. search a registered object with a given name */
	if( ubus_lookup_id(ctx, "ering.dmount", &host_id) ) {
		printf("ering.dmount is NOT found in ubus!\n");
		goto UBUS_FAIL;
	}
	printf("ering.dmount is found in ubus @%u.\n",host_id);
	
	/* 6. prepare request method policy and data */
	blob_buf_init(&bb_t,0);
	blobmsg_add_u32(&bb_t,"id", cli_obj.id); 	/* ID */
	blobmsg_add_string(&bb_t,"device", arr2[2]);   /* DEVICE */

	switch (switch_item) {
		case opMount:
		   blobmsg_add_string(&bb_t,"mountpoint", argv[3]);   /* MTPT */
		   ret=ubus_invoke(ctx, host_id, "mount", bb_t.head, result_handler, 0, 3000);
           break;
		case opUmount:
		   ret=ubus_invoke(ctx, host_id, "umount", bb_t.head, result_handler, 0, 3000);
		   break;
		case opRemove:
		   ret=ubus_invoke(ctx, host_id, "remove", bb_t.head, result_handler, 0, 3000);
           break;
           
		default:
		   break;
	}
 
	/* 7. call the ubus host object */
	
	printf("Call result: %s\n", ubus_strerror(ret));
 
	/* 8. uloop routine: events monitoring and callback provoking
	      However, we just ignore uloop in this case.
	 */
	 //uloop_run();
 
	uloop_done();
	
UBUS_FAIL:
	ubus_free(ctx);
	
	return 0;
}

/**
 * \brief Callback function for ubus_invoke to process result from the host
 * Here we just print out the message.
 */
static void result_handler(struct ubus_request *req, int type, struct blob_attr *msg_t)
{
        char *strmsg;
 
        if(!msg_t)
                return;
 
        strmsg=blobmsg_format_json_indent(msg_t,true, 0); /* 0 type of format */
        printf("Response from the host: %s\n", strmsg);
        free(strmsg); /* need to free strmsg */
}
