 /*
  * server.c
  * 
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
  * Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  * 
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  * See the GNU General Public License for more details.
  * 
  * You should have received a copy of the GNU General Public License along
  * with this program.  If not, see <http://www.gnu.org/licenses/>.
  * 
  * See the COPYING file.
  */


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <syslog.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <libgen.h>     // dirname() and basename()
#include <ctype.h>      // isdigit()	  
#include <errno.h>

#include <libubus.h>

#include "def.h"
#include "init_list.h"

#define MAX_SIZE 256

#define strdup_or_null( str )  (str) != NULL ? strdup(str) : NULL

const char *progname="dmount";
const char *pid_file="/var/run/dmountd.pid";
static bool foreground = false;
static int pid_fd = -1;

void parse_args(int argc, char *argv[]);
pid_t get_pid_t(void);
int wait_on_kill(int s, int m);
void daemonize();
 
static struct ubus_context *ctx;
static struct ubus_request_data req_data;
static struct blob_buf bb_t;

/**
 * \brief Callback function for handling signals.
 * \param	sig	identifier of signal
 */
static void handle_signal(int sig)
{
	if (sig == SIGINT || sig == SIGTERM || sig == SIGQUIT) {
		
		syslog(LOG_INFO, "Debug: stopping daemon ...\n");
		
		if (pid_fd != -1) {
			int rc = lockf(pid_fd, F_ULOCK, 0);
			if(rc != 0) exit(EXIT_FAILURE);
			close(pid_fd);
		}
		
		if (pid_file) {
			unlink(pid_file);
		}
		
		/* Reset signal handling to default behavior */
		signal(SIGINT,  SIG_DFL);
		signal(SIGTERM, SIG_DFL);
		signal(SIGQUIT, SIG_DFL);
	}
}
 
/* Enum for EGI policy order */
enum {
	MOUNT_ID,
	MOUNT_DEVICE,
	MOUNT_MTPT,
	__MOUNT_MAX,
};
 
/* Ubus Policy */
static const struct blobmsg_policy mount_policy[] =
{
	[MOUNT_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32},
	[MOUNT_DEVICE] = { .name="device", .type=BLOBMSG_TYPE_STRING },
	[MOUNT_MTPT] = { .name="mountpoint", .type=BLOBMSG_TYPE_STRING },
};
 
/* Enum for EGI policy order */
enum {
	UMOUNT_ID,
	UMOUNT_DEVICE,
	__UMOUNT_MAX,
};
 
/* Ubus Policy */
static const struct blobmsg_policy umount_policy[] =
{
	[UMOUNT_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32},
	[UMOUNT_DEVICE] = { .name="device", .type=BLOBMSG_TYPE_STRING },
};
 
/* Enum for EGI policy order */
enum {
	REMOVE_ID,
	REMOVE_DEVICE,
	__REMOVE_MAX,
};
 
/* Ubus Policy */
static const struct blobmsg_policy remove_policy[] =
{
	[REMOVE_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32},
	[REMOVE_DEVICE] = { .name="device", .type=BLOBMSG_TYPE_STRING },
};

/**
 * \brief Callback function for ubus mount method handling
 */
static int mount_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
	char cmd[64] = {0};
	int rc = 0;
	char target[1024]={0};
    struct nodeList *l = NULL;
    bool found = false;
	
	/* for parsed attr */
	struct blob_attr *tb[__MOUNT_MAX];
	 
	/* Parse blob_msg from the caller to request policy */
	blobmsg_parse(mount_policy, ARRAY_SIZE(mount_policy), tb, blob_data(msg_t), blob_len(msg_t));
   
    l = get_devices(l);
   
    struct nodeList *aux = l;
    while(aux) {
       char* tmp = rindex( aux->nodeT.sysfs_path, '/' );
	   if(!strcmp(blobmsg_data(tb[REMOVE_DEVICE]), tmp+1)) {
		   found = true;
		   break;
       }
       
       aux = aux->next;
    }
        
    if(l) 
       l = clearList(l);
       
    if(!found) return -1;	      
	
    rc = mkdir(blobmsg_data(tb[MOUNT_MTPT]), 0777);
    if (rc == -1) {
        switch (errno) {
            case EACCES :
                log_error("%s: the parent directory does not allow write\n", progname);
                exit(EXIT_FAILURE);
            case EEXIST:
                log_info("%s: pathname already exists\n", progname);
                break;
            case ENAMETOOLONG:
                log_error("%s: pathname is too long\n", progname);
                exit(EXIT_FAILURE);
            default:
                perror("mkdir");
                exit(EXIT_FAILURE);
        }
    }
    	
    sprintf(target, blobmsg_data(tb[MOUNT_MTPT]));
    strcat(target, "/");
    strcat(target, blobmsg_data(tb[MOUNT_DEVICE]));
    	
    rc = mkdir(target, 0777);
    if (rc == -1) {
        switch (errno) {
            case EACCES :
                log_error("%s: the parent directory does not allow write\n", progname);
                exit(EXIT_FAILURE);
            case EEXIST:
                log_info("%s: pathname already exists\n", progname);
                break;
            case ENAMETOOLONG:
                log_error("%s: pathname is too long\n", progname);
                exit(EXIT_FAILURE);
            default:
                perror("mkdir");
                exit(EXIT_FAILURE);
        }
    }
	
    sprintf(cmd, "mount --rw /dev/");
    strcat(cmd, blobmsg_data(tb[MOUNT_DEVICE]));
    strcat(cmd, " ");
    strcat(cmd, target);
 
	/* Do the job here according to caller's request */
	system(cmd);
 
	/* send a reply msg to the caller for information */
	blob_buf_init(&bb_t, 0);
	blobmsg_add_string(&bb_t,"Server reply - Request is being proceeded: ", cmd);
	ubus_send_reply(ctx, req, bb_t.head);
 
	/* 	-----  reply results to the caller ----- */
	ubus_defer_request(ctx, req, &req_data);
	ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
	
	return 0;
}

/**
 * \brief Callback function for ubus umount method handling
 */
static int umount_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
	char cmd[64] = {0};
    struct nodeList *l = NULL;
    bool found = false;
	
	/* for parsed attr */
	struct blob_attr *tb[__UMOUNT_MAX]; 
 
	/* Parse blob_msg from the caller to request policy */
	blobmsg_parse(umount_policy, ARRAY_SIZE(umount_policy), tb, blob_data(msg_t), blob_len(msg_t));
   
    l = get_devices(l);
   
    struct nodeList *aux = l;
    while(aux) {
       char* tmp = rindex( aux->nodeT.sysfs_path, '/' );
	   if(!strcmp(blobmsg_data(tb[REMOVE_DEVICE]), tmp+1)) {
		   found = true;
		   break;
       }
       
       aux = aux->next;
    }
        
    if(l) 
       l = clearList(l);
       
    if(!found) return -1;	      
	
    sprintf(cmd, "umount /dev/");
    strcat(cmd, blobmsg_data(tb[UMOUNT_DEVICE]));
 
	/* Do the job here according to caller's request */
	system(cmd);
 
	/* send a reply msg to the caller for information */
	blob_buf_init(&bb_t, 0);
	blobmsg_add_string(&bb_t,"Server reply - Request is being proceeded: ", cmd);
	ubus_send_reply(ctx, req, bb_t.head);
 
	/* 	-----  reply results to the caller ----- */
	ubus_defer_request(ctx, req, &req_data);
	ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
	
	return 0;
}

/**
 * \brief Callback function for ubus remove method handling
 */
static int remove_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
	char cmd[64] = {0};
	char device[64]={0};
    struct nodeList *l = NULL;
    char line[300];
    FILE *pf = NULL;
    bool unmounted = true;
    int len = 0;
	
	/* for parsed attr */
	struct blob_attr *tb[__REMOVE_MAX]; 
 
	/* Parse blob_msg from the caller to request policy */
	blobmsg_parse(remove_policy, ARRAY_SIZE(remove_policy), tb, blob_data(msg_t), blob_len(msg_t));
   
    l = get_devices(l);
   
    struct nodeList *aux = l;
    while(aux) {
       char* tmp = rindex( aux->nodeT.sysfs_path, '/' );
	   if(!strcmp(blobmsg_data(tb[REMOVE_DEVICE]), tmp+1)) {
		   sprintf(device, blobmsg_data(tb[REMOVE_DEVICE]));
		   if(aux->nodeT.removable) {
	          sprintf(cmd, "/usr/bin/eject /dev/");
			  if(!strcmp(aux->nodeT.devtype_str, "partition")) {
			     int len = strlen(device);
			     device[len-1]='\0';
                 strcat(cmd, device);
              } else if(!strcmp(aux->nodeT.devtype_str, "disk")) {
                 strcat(cmd, device);
              }
	       } else {
              sprintf(cmd, "echo 1 | tee /sys/block/");
			  if(!strcmp(aux->nodeT.devtype_str, "partition")) {
			     int len = strlen(device);
			     device[len-1]='\0';
                 strcat(cmd, device);
                 strcat(cmd, "/device/delete");
              } else if(!strcmp(aux->nodeT.devtype_str, "disk")) {
                 strcat(cmd, device);
                 strcat(cmd, "/device/delete");
              }
           }
           break;
       }
       
       aux = aux->next;
    }
            
    if(l) 
       l = clearList(l);
       
    pf=fopen("/proc/self/mountinfo", "r");
    if(!pf) {
	   log_error("%s: Pipe error\n", progname);
	   exit(EXIT_FAILURE);
	}  
	    
    /* decode every line of /proc/self/mountinfo */
    while( fgets(line, sizeof(line), pf) ) {
	   int mp[2] = {0,0}, fs[2], df[2] = {0,0};
	   char *mtpt, *fstp, *devf, *partitionname;
	   /* get start and end of mountpoint path string and device file name */
	   sscanf(line, "%*s %*s %*s %*s %n%*s%n%*[^-]- %n%*s%n %n%*s%n",
	          mp, mp+1, fs, fs+1, df, df+1);
	   mtpt = line+mp[0];
	   fstp = line+fs[0];
	   devf = line+df[0];
	   line[mp[1]] = line[fs[1]] = line[df[1]] = '\0';
	   partitionname = basename(devf);
	   len=strlen(partitionname);
	   if(isdigit(partitionname[len-1])) partitionname[len-1] = '\0';
	   if(!strcmp(partitionname, device)) {
	      unmounted = false;
	      break;
	   }
    }
    
    fclose(pf);
       
    if(!unmounted) return -1;
    
	/* Do the job here according to caller's request */
	system(cmd);
 
	/* send a reply msg to the caller for information */
	blob_buf_init(&bb_t, 0);
	blobmsg_add_string(&bb_t,"Server reply - Request is being proceeded: ", cmd);
	ubus_send_reply(ctx, req, bb_t.head);
 
	/* 	-----  reply results to the caller ----- */
	ubus_defer_request(ctx, req, &req_data);
	ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
	
	return 0;
}
 
/* Ubus Methods */
static const struct ubus_method dmount_methods[] =
{
	UBUS_METHOD("mount", mount_handler, mount_policy),
	UBUS_METHOD("umount", umount_handler, umount_policy),
	UBUS_METHOD("remove", remove_handler, remove_policy),
};
 
/* Ubus object type */
static struct ubus_object_type dmount_obj_type =
	UBUS_OBJECT_TYPE("ering.dmount", dmount_methods);
 
/* Ubus object */
static struct ubus_object dmount_obj=
{
	.name = "ering.dmount", 	/* with APP name */
	.type = &dmount_obj_type,
	.methods = dmount_methods,
	.n_methods = ARRAY_SIZE(dmount_methods),
	//.path= /* useless */
};
 
int main(int argc, char **argv)
{      
   int ret;
   const char *ubus_socket="/var/run/ubus/ubus.sock";    
   FILE *fp = NULL; 
   int count=0;
	
   parse_args(argc, argv);

   fp=popen("ps -A | grep \"dmountd$\"", "r");
   if(!fp) {
      printf("Sorry, there was an error opening the pipe\n");
      exit(EXIT_FAILURE);
   }

   char line[MAX_SIZE]={0};
   while (fgets(line, MAX_SIZE, fp)) count++;
   pclose(fp);
   
   if(count > 2) {
      printf("Too many instances of %s\n", progname);
	  exit(EXIT_FAILURE);
   }
	
	if (foreground == 0) {
		
		daemonize();
	}

	/* Open system log and write message to it */
	openlog(argv[0], LOG_PID|LOG_CONS, LOG_DAEMON);
	syslog(LOG_INFO, "Started %s", progname);
	
	signal(SIGINT, handle_signal);
	signal(SIGQUIT, handle_signal);
	signal(SIGTERM, handle_signal);
	
	/* 1. create an epoll instance descriptor poll_fd */
	uloop_init();
 
	/* 2. connect to ubusd and get ctx */
	ctx=ubus_connect(ubus_socket);
	if(ctx==NULL) exit(EXIT_FAILURE);
 
	/* 3. registger epoll events to uloop, start sock listening */
	ubus_add_uloop(ctx);
 
	/* 4. register a usb_object to ubusd */
	ret=ubus_add_object(ctx, &dmount_obj);
	if(ret!=0) goto UBUS_FAIL;
 
	/* 5. uloop routine: events monitoring and callback provoking */
	uloop_run(); 
	uloop_done();
	
UBUS_FAIL:
	ubus_free(ctx);
	
	syslog(LOG_INFO, "Stopped %s", progname);
	closelog();
	
	return 0;
}

void parse_args(int argc, char *argv[])
{
   static struct option long_options[] = {
	   
      {"foreground",            no_argument, 0, 'f'},
      {"kill",                  no_argument, 0, 'k'},
      {"check-running",         no_argument, 0, 'c'},
      {0, 0, 0}
   };
   
   int rc = 0;
   int value, option_index = 0;
   int _kill = 0, _check = 0;
   while ((value = getopt_long(argc, argv, "fkc", long_options, &option_index)) != -1) {
      switch (value) {
         case 0:
            if (long_options[option_index].flag)
            break;
            
         case 'f':
            foreground = true;
            break;
                
         case 'k':
            _kill = 1;
            break;
            
         case 'c':
            _check = 1;
            break;
                  
         default:
            syslog(LOG_ERR, "Unknown parameter.");
            rc = -1;
            break;
      }
      
   }
   
   if (_kill) {
      
      if ((rc = wait_on_kill(SIGTERM, 60)) < 0) {
         printf("Failed to kill daemon: %s with SIGTERM , rc = %d\n", strerror(errno), rc);
         exit(1);
      }
      
      exit(EXIT_SUCCESS);
   }
   
   if (_check) {
      
      pid_t pid = get_pid_t();
      
      if (pid == (pid_t) -1 || pid == 0)
         printf("%s not running.\n", progname);
      else
         printf("%s process running as pid %u.\n", progname, pid);
      
      exit(EXIT_SUCCESS);
   }
}

pid_t get_pid_t(void)
{
   static char txt[MAX_SIZE];
   int fd = -1;
   pid_t ret = (pid_t) -1, pid;
   ssize_t l;
   long lpid;
   char *e = NULL;
   
   if ((fd = open(pid_file, O_RDWR, 0644)) < 0) {
      if ((fd = open(pid_file, O_RDONLY, 0644)) < 0) {
         if (errno != ENOENT) {
			 syslog(LOG_PERROR, "Failed to open pidfile '%s': %s , errno=%d\n", pid_file, strerror(errno), errno);
         }
         goto finish;
      }
   }
   
   if ((l = read(fd, txt, MAX_SIZE-1)) < 0) {
      int saved_errno = errno;
      syslog(LOG_PERROR, "read(): %s , errno=%d", strerror(errno), errno);
      unlink(pid_file);
      errno = saved_errno;
      goto finish;
   }
   
   txt[l] = 0;
   txt[strcspn(txt, "\r\n")] = 0;
   
   errno = 0;
   lpid = strtol(txt, &e, 10);
   pid = (pid_t) lpid;
   
   if (errno != 0 || !e || *e || (long) pid != lpid) {
      unlink(pid_file);
      errno = EINVAL;
      syslog(LOG_WARNING, "PID file corrupt, removing. (%s) , errno=%d", pid_file, errno);
      goto finish;
   }
    
   if (kill(pid, 0) != 0 && errno != EPERM) {
      syslog(LOG_WARNING, "Process %lu died: %s; trying to remove PID file. (%s)", (unsigned long) pid, strerror(errno), pid_file);
      unlink(pid_file);
      goto finish;
   }
   
   ret = pid;

finish:
   if (fd != -1) {
      lockf(fd, F_ULOCK, 0);
	  close(fd);
   }
   
   return ret;
}

int wait_on_kill(int s, int m)
{
   pid_t pid;
   time_t t;

   if ((pid = get_pid_t()) < 0) 
      return -1;
      
   if (kill(pid, s) < 0)
      return -1;

   t = time(NULL) + m;

   for (;;) {
      int r;
      struct timeval tv = { 0, 100000 };

      if (time(NULL) > t) {
         errno = ETIME;
         return -1;
      }

      if ((r = kill(pid, 0)) < 0 && errno != ESRCH)
         return -1;

      if (r)
         return 0;
         
      kill(pid, SIGKILL);
      
      if (select(0, NULL, NULL, NULL, &tv) < 0)
         return -1;
    }
}

void daemonize()
{
   pid_t pid = 0;
   
   pid = fork();
   
   if (pid < 0) {
      exit(EXIT_FAILURE);
   }
   
   if (pid > 0) {
      exit(EXIT_SUCCESS);
   }

   if (setsid() < 0) {
      exit(EXIT_FAILURE);
   }
   
   signal(SIGCHLD, SIG_IGN);
   
   pid = fork();
   
   if (pid < 0) {
      exit(EXIT_FAILURE);
   }
   
   if (pid > 0) {
      exit(EXIT_SUCCESS);
   }
   
   umask(0);
   
   chdir("/tmp");
   
   stdin  = fopen("/dev/null", "r");
   stdout = fopen("/dev/null", "w+");
   stderr = fopen("/dev/null", "w+");
   
   if (pid_file)
   {
      char str[MAX_SIZE];
      pid_fd = open(pid_file, O_RDWR|O_CREAT, 0640);
      if (pid_fd < 0) {
         exit(EXIT_FAILURE);
      }
      
      if (lockf(pid_fd, F_TLOCK, 0) < 0) {
         exit(EXIT_FAILURE);
      }
      
      sprintf(str, "%d\n", getpid());
      
      if (write(pid_fd, str, strlen(str)) < 0) {
         exit(EXIT_FAILURE);
      }
   }
}
